@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* exec.c: command execution */
-/* (C) 2007-2024 magicant */
+/* (C) 2007-2025 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -665,9 +665,8 @@ void exec_simple_command(const command_T *c, bool finally_exit)
     lastcmdsubstatus = Exit_SUCCESS;
 
     /* expand the command words */
-    int argc;
-    void **argv;
-    if (!expand_line(c->c_words, &argc, &argv)) {
+    plist_T args = expand_line(c->c_words, c->c_isdeclutil);
+    if (args.contents == NULL) {
         laststatus = Exit_EXPERROR;
         goto done;
     }
@@ -675,15 +674,15 @@ void exec_simple_command(const command_T *c, bool finally_exit)
         goto done1;
 
     /* execute the remaining part */
-    if (argc == 0)
+    if (args.length == 0)
         finally_exit |= exec_simple_command_without_words(c);
     else
-        finally_exit |=
-            exec_simple_command_with_words(c, argc, argv, finally_exit);
+        finally_exit |= exec_simple_command_with_words(
+                c, args.length, args.contents, finally_exit);
 
     /* cleanup */
 done1:
-    plfree(argv, free);
+    plfree(pl_toary(&args), free);
 done:
     if (finally_exit)
         /* If we're running the EXIT trap and the simple command failed with a
@@ -1286,22 +1285,25 @@ void exec_for(const command_T *c, bool finally_exit)
     execstate.loopnest++;
     execstate.breakloopnest = execstate.loopnest;
 
-    int count;
+    size_t count;
     void **words;
 
     if (c->c_forwords != NULL) {
         /* expand the words between "in" and "do" of the for command. */
-        if (!expand_line(c->c_forwords, &count, &words)) {
+        plist_T wordlist = expand_line(c->c_forwords, false);
+        if (wordlist.contents == NULL) {
             laststatus = Exit_EXPERROR;
             apply_errexit_errreturn(NULL);
             goto finish;
         }
+        count = wordlist.length;
+        words = pl_toary(&wordlist);
     } else {
         /* no "in" keyword in the for command: use the positional parameters */
         struct get_variable_T v = get_variable(L"@");
         assert(v.type == GV_ARRAY && v.values != NULL);
         save_get_variable_values(&v);
-        count = (int) v.count;
+        count = v.count;
         words = v.values;
     }
 
@@ -1317,18 +1319,16 @@ void exec_for(const command_T *c, bool finally_exit)
         goto done;                                      \
     } else (void) 0
 
-    int i;
+    size_t i;
+    bool assignsuccess = true;
     for (i = 0; i < count; i++) {
-        if (!set_variable(c->c_forname, words[i],
-                    shopt_forlocal && !posixly_correct ?
-                        SCOPE_LOCAL : SCOPE_GLOBAL,
-                    false)) {
-            laststatus = Exit_ASSGNERR;
-            apply_errexit_errreturn(NULL);
-            if (!is_interactive_now)
-                finally_exit = true;
-            goto done;
-        }
+        assignsuccess = set_variable(
+                c->c_forname, words[i],
+                shopt_forlocal && !posixly_correct ? SCOPE_LOCAL : SCOPE_GLOBAL,
+                false);
+        if (!assignsuccess)
+            break;
+
         exec_and_or_lists(c->c_forcmds, finally_exit && i + 1 == count);
 
         if (c->c_forcmds == NULL)
@@ -1340,6 +1340,14 @@ done:
     while (++i < count)  /* free unused words */
         free(words[i]);
     free(words);
+
+    if (!assignsuccess) {
+        laststatus = Exit_ASSGNERR;
+        apply_errexit_errreturn(NULL);
+        if (!is_interactive_now)
+            finally_exit = true;
+    }
+
     if (count == 0 && c->c_forcmds != NULL)
         laststatus = Exit_SUCCESS;
 finish:
@@ -1395,37 +1403,52 @@ void exec_case(const command_T *c, bool finally_exit)
     if (word == NULL)
         goto fail;
 
+    bool match = false, emptycmd = true;
+
     for (const caseitem_T *ci = c->c_casitems; ci != NULL; ci = ci->next) {
-        for (void **pats = ci->ci_patterns; *pats != NULL; pats++) {
+        for (void **pats = ci->ci_patterns; !match && *pats != NULL; pats++) {
             wchar_t *pattern =
                 expand_single(*pats, TT_SINGLE, Q_WORD, ES_QUOTED);
             if (pattern == NULL)
                 goto fail;
 
-            bool match = match_pattern(word, pattern);
+            match = match_pattern(word, pattern);
             free(pattern);
-            if (match) {
-                if (ci->ci_commands != NULL) {
-                    exec_and_or_lists(ci->ci_commands, finally_exit);
-                    goto done;
-                } else {
-                    goto success;
-                }
-            }
+        }
+        if (!match)
+            continue;
+
+        exec_and_or_lists(
+                ci->ci_commands,
+                finally_exit && (ci->next == NULL || ci->ci_cont == CC_BREAK));
+        emptycmd = (ci->ci_commands == NULL);
+
+        switch (ci->ci_cont) {
+            case CC_BREAK:
+                goto done;
+            case CC_FALLTHRU:
+                match = true;
+                break;
+            case CC_CONTINUE:
+                match = false;
+                break;
         }
     }
-success:
-    laststatus = Exit_SUCCESS;
+
 done:
+    if (emptycmd)
+        laststatus = Exit_SUCCESS;
     free(word);
     if (finally_exit)
         exit_shell();
     return;
 
 fail:
+    free(word);
     laststatus = Exit_EXPERROR;
     apply_errexit_errreturn(NULL);
-    goto done;
+    if (finally_exit)
+        exit_shell();
 }
 
 /* Executes the function definition. */
@@ -1630,7 +1653,7 @@ wchar_t *exec_command_substitution(const embedcmd_T *cmdsub)
         xclose(pipefd[PIPE_IN]);
         if (pipefd[PIPE_OUT] != STDOUT_FILENO) {  /* connect the pipe */
             if (xdup2(pipefd[PIPE_OUT], STDOUT_FILENO) < 0)
-                exit(Exit_NOEXEC);
+                _Exit(Exit_NOEXEC);
             xclose(pipefd[PIPE_OUT]);
         }
 
@@ -2298,8 +2321,8 @@ error3:
 error2:
     free(mbssaveargv0);
 error1:
-    if (posixly_correct || !is_interactive_now)
-        exit(err);
+    if (!is_interactive_now)
+        _Exit(err);
     return err;
 }
 
