@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* expand.c: word expansion */
-/* (C) 2007-2021 magicant */
+/* (C) 2007-2025 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -64,7 +64,7 @@ static inline void fill_ccbuf(
     __attribute__((nonnull));
 
 static wchar_t *expand_tilde(const wchar_t **ss,
-        bool hasnextwordunit, tildetype_T tt)
+        bool hasnextwordunit, bool stopatcolon)
     __attribute__((nonnull,malloc,warn_unused_result));
 
 enum indextype_T { IDX_NONE, IDX_ALL, IDX_CONCAT, IDX_NUMBER, };
@@ -140,6 +140,8 @@ static void add_empty_field(plist_T *dest, const wchar_t *p)
 static inline void add_sq(
         const wchar_t *restrict *ss, xwcsbuf_T *restrict buf, bool escape)
     __attribute__((nonnull));
+static wchar_t *interpret_dsq(const wchar_t *restrict *ss)
+    __attribute__((nonnull,malloc,warn_unused_result));
 static inline bool should_escape(charcategory_T cc, escaping_T escaping)
     __attribute__((const));
 static wchar_t *quote_removal_free(
@@ -160,34 +162,44 @@ static void maybe_exit_on_error(void);
 /* Expands a command line.
  * `args' is a NULL-terminated array of pointers to `const wordunit_T'
  * to expand.
- * If successful, the number of resulting words is assigned to `*argcp', a
- * pointer to a newly malloced array of the expanded words is assigned to
- * `*argvp', and true is returned. The array is NULL-terminated and its elements
- * are newly malloced wide strings.
- * If unsuccessful, false is returned and the values of `*argcp' and `*argvp'
- * are indeterminate.
+ * If `assignsingle' is true, assignment-like words are expanded to single
+ * fields like expanding assignment tokens. Otherwise, all words are expanded to
+ * multiple fields as usual.
+ * If successful, a list of pointers to newly malloced wide strings is returned.
+ * On error, a non-initialized pointer list is returned whose `contents' field
+ * is NULL.
  * On error in a non-interactive shell, the shell exits. */
-bool expand_line(void *const *restrict args,
-    int *restrict argcp, void ***restrict argvp)
+plist_T expand_line(void *const *args, bool assignsingle)
 {
     plist_T list;
     pl_init(&list);
 
     for (; *args != NULL; args++) {
-        if (!expand_multiple(*args, &list)) {
+        const wordunit_T *w = *args;
+        bool success;
+        if (!assignsingle || w == NULL || w->wu_type != WT_STRING ||
+                !is_assignment_prefix(w->wu_string)) {
+            success = expand_multiple(w, &list);
+        } else {
+            wchar_t *s = expand_single(w, TT_ASSIGN, Q_WORD, ES_NONE);
+            success = (s != NULL);
+            if (success)
+                pl_add(&list, s);
+        }
+        if (!success) {
             plfree(pl_toary(&list), free);
-            return false;
+            list.contents = NULL;
+            break;
         }
     }
 
-    *argcp = list.length;
-    *argvp = pl_toary(&list);
-    return true;
+    return list;
 }
 
-/* Expands a word.
- * The results, which are added to `list' as newly-malloced wide strings, may
- * be multiple words.
+/* Expands a word to (possibly any number of) fields.
+ * The operation includes the four expansions, brace expansion, field splitting,
+ * pathname expansion, and quote removal.
+ * The results are added to `list' as newly-malloced wide strings.
  * The return value is true iff successful.
  * On error in a non-interactive shell, the shell exits. */
 bool expand_multiple(const wordunit_T *w, plist_T *list)
@@ -226,6 +238,31 @@ bool expand_multiple(const wordunit_T *w, plist_T *list)
     return true;
 }
 
+/* Expands a word to (possibly any number of) fields.
+ * The operation includes the four expansions and quote removal. Unlike
+ * `expand_multiple', this function does not perform brace expansion, field
+ * splitting, or pathname expansion.
+ * If successful, the return value is a plist_T containing newly malloced wide
+ * strings. In most cases, the plist_T contains one string. If the word contains
+ * "$@", however, it may contain any number of strings.
+ * On error, the return value is a plist_T with `contents' being NULL. */
+plist_T expand_word(const wordunit_T *w)
+{
+    /* four expansions */
+    struct expand_four_T expand = expand_four(w, TT_NONE, Q_WORD, CC_LITERAL);
+    if (expand.valuelist.contents == NULL)
+        return expand.valuelist;
+
+    /* quote removal */
+    for (size_t i = 0; i < expand.valuelist.length; i++)
+        expand.valuelist.contents[i] = quote_removal_free(
+            expand.valuelist.contents[i], expand.cclist.contents[i], ES_NONE);
+
+    pl_destroy(&expand.cclist);
+
+    return expand.valuelist;
+}
+
 /* Expands a word to a single field.
  * If successful, the result is a pair of newly malloced strings.
  * On error, an error message is printed and a NULL pair is returned.
@@ -274,26 +311,6 @@ struct cc_word_T expand_single_cc(
     return (struct cc_word_T) { wb_towcs(&valuebuf), sb_tostr(&ccbuf) };
 }
 
-/* Expands a word to (possibly any number of) fields.
- * If successful, the return value is a plist_T containing newly malloced wide
- * strings. In most cases, the plist_T contains one string. If the word contains
- * "$@", however, it may contain any number of strings.
- * On error, the return value is a plist_T with `contents' being NULL. */
-plist_T expand_word(const wordunit_T *w)
-{
-    /* four expansions */
-    struct expand_four_T expand = expand_four(w, TT_NONE, Q_WORD, CC_LITERAL);
-
-    /* quote removal */
-    for (size_t i = 0; i < expand.valuelist.length; i++)
-        expand.valuelist.contents[i] = quote_removal_free(
-            expand.valuelist.contents[i], expand.cclist.contents[i], ES_NONE);
-
-    pl_destroy(&expand.cclist);
-
-    return expand.valuelist;
-}
-
 /* Expands a single word: the four expansions and quote removal.
  * This function doesn't perform brace expansion, field splitting, or globbing.
  * If successful, the resulting word is returned as a newly malloced string.
@@ -426,8 +443,8 @@ struct expand_four_T expand_four(const wordunit_T *restrict w,
         switch (w->wu_type) {
         case WT_STRING:
             ss = w->wu_string;
-            if (first && tilde != TT_NONE) {
-                s = expand_tilde(&ss, w->next, tilde);
+            if (first && (tilde == TT_SINGLE || tilde == TT_MULTI)) {
+                s = expand_tilde(&ss, w->next != NULL, tilde == TT_MULTI);
                 if (s != NULL) {
                     wb_catfree(&valuebuf, s);
                     fill_ccbuf(&valuebuf, &ccbuf,
@@ -472,6 +489,23 @@ struct expand_four_T expand_four(const wordunit_T *restrict w,
                     assert(*ss == L'\'');
                     fill_ccbuf(&valuebuf, &ccbuf, defaultcc | CC_QUOTED);
 
+                    wb_wccat(&valuebuf, L'\'');
+                    sb_ccat(&ccbuf, defaultcc | CC_QUOTATION);
+                    break;
+                case L'$':
+                    // Check for dollar-single-quotes.
+                    if (quoting != Q_WORD || indq || ss[1] != '\'')
+                        goto default_;
+
+                    wb_wccat(&valuebuf, L'$');
+                    wb_wccat(&valuebuf, L'\'');
+                    sb_ccat(&ccbuf, defaultcc | CC_QUOTATION);
+                    sb_ccat(&ccbuf, defaultcc | CC_QUOTATION);
+
+                    wb_catfree(&valuebuf, interpret_dsq(&ss));
+                    assert(*ss == L'\'');
+                    fill_ccbuf(&valuebuf, &ccbuf, defaultcc | CC_QUOTED);
+
                     wb_wccat(&valuebuf, L'\'');
                     sb_ccat(&ccbuf, defaultcc | CC_QUOTATION);
                     break;
@@ -504,15 +538,22 @@ struct expand_four_T expand_four(const wordunit_T *restrict w,
                 case L':':
                     if (indq || tilde != TT_MULTI)
                         goto default_;
-
-                    /* perform tilde expansion after a colon */
-                    wb_wccat(&valuebuf, L':');
+                    /* perform tilde expansion after the colon */
+                    goto tilde;
+                case L'=':
+                    if (indq || tilde != TT_ASSIGN)
+                        goto default_;
+                    tilde = TT_MULTI;
+                    /* perform tilde expansion after the equal */
+tilde:
+                    wb_wccat(&valuebuf, *ss);
                     sb_ccat(&ccbuf, defaultcc);
                     ss++;
-                    s = expand_tilde(&ss, w->next, tilde);
+                    s = expand_tilde(&ss, w->next != NULL, true);
                     if (s != NULL) {
                         wb_catfree(&valuebuf, s);
-                        fill_ccbuf(&valuebuf, &ccbuf, CC_HARD_EXPANSION);
+                        fill_ccbuf(&valuebuf, &ccbuf,
+                                CC_HARD_EXPANSION | (defaultcc & CC_QUOTED));
                     }
                     continue;
 default_:
@@ -591,14 +632,15 @@ void fill_ccbuf(const xwcsbuf_T *restrict valuebuf, xstrbuf_T *restrict ccbuf,
  * If `**ss' is not L'~' or expansion fails, this function has no side effects
  * and returns NULL. If successful, `*ss' is incremented and the result is
  * returned as a newly malloced string. */
-wchar_t *expand_tilde(const wchar_t **ss, bool hasnextwordunit, tildetype_T tt)
+wchar_t *expand_tilde(
+        const wchar_t **ss, bool hasnextwordunit, bool stopatcolon)
 {
     const wchar_t *s = *ss;
     if (*s != L'~')
         return NULL;
     s++;
 
-    const wchar_t *end = wcspbrk(s, tt == TT_SINGLE ? L"/" : L"/:");
+    const wchar_t *end = wcspbrk(s, stopatcolon ? L"/:" : L"/");
     wchar_t *username;
     const wchar_t *home;
     size_t usernamelen;
@@ -1702,6 +1744,119 @@ void add_sq(const wchar_t *restrict *ss, xwcsbuf_T *restrict buf, bool escape)
     }
 }
 
+/* Expands the content of a dollar-single-quoted string to a newly-malloced
+ * string.
+ * `ss' is a pointer to a pointer to the string to be expanded. Initially,
+ * `(*ss)[0]' and `(*ss)[1]' must be '$' and '\'', respectively, which are
+ * skipped in this function. The following characters in the string are
+ * accumulated in the string to be returned, interpreting any backslash escapes
+ * encountered, until a closing '\'' is found (or the end of the string is
+ * reached). When this function returns, `*ss' is updated so that `**ss' is the
+ * closing quote or terminating null character. */
+/* If an escape contained in `*ss' produces a null character, the rest of the
+ * string is ignored by the caller. This is one of the behaviors allowed by
+ * POSIX. This function could have been designed to take an `xwcsbuf_T *'
+ * argument and append the result directly to it, but that would not be
+ * compliant because the null character would also hide the characters after the
+ * closing quote. */
+wchar_t *interpret_dsq(const wchar_t *restrict *ss)
+{
+    xwcsbuf_T buf;
+    wb_init(&buf);
+
+    const wchar_t *s = *ss;
+    assert(s[0] == '$');
+    assert(s[1] == '\'');
+    s += 2;
+
+    for (;;) {
+        switch (*s) {
+            case L'\0':
+            case L'\'':
+                *ss = s;
+                return wb_towcs(&buf);
+            case L'\\':
+                s++;
+                switch (*s) {
+                    case L'"':
+                    case L'\'':
+                    case L'\\':
+                        wb_wccat(&buf, *s);
+                        s++;
+                        break;
+                    case L'a':  wb_wccat(&buf, L'\a');    s++;  break;
+                    case L'b':  wb_wccat(&buf, L'\b');    s++;  break;
+                    case L'e':  wb_wccat(&buf, L'\033');  s++;  break;
+                    case L'f':  wb_wccat(&buf, L'\f');    s++;  break;
+                    case L'n':  wb_wccat(&buf, L'\n');    s++;  break;
+                    case L'r':  wb_wccat(&buf, L'\r');    s++;  break;
+                    case L't':  wb_wccat(&buf, L'\t');    s++;  break;
+                    case L'v':  wb_wccat(&buf, L'\v');    s++;  break;
+                    case L'c':
+                        s++;
+                        wchar_t c;
+                        if (*s == L'\\') {
+                            s++;
+                            if (*s == L'\\')
+                                c = L'\\' ^ 0x40;
+                            else // Oops, unknown escape!
+                                c = L'?';
+                        } else if (*s == L'?' || (L'A' <= *s && *s <= L'_')) {
+                            c = *s ^ 0x40;
+                        } else if (L'a' <= *s && *s <= L'z') {
+                            c = *s ^ 0x60;
+                        } else { // Oops, unknown escape!
+                            c = L'?';
+                        }
+                        wb_wccat(&buf, c);
+                        s++;
+                        break;
+                    case L'x':
+                        s++;
+                        int value;
+                        if (L'0' <= *s && *s <= L'9') {
+                            value = *s - L'0';
+                        } else if (L'A' <= *s && *s <= L'F') {
+                            value = *s - L'A' + 0xA;
+                        } else if (L'a' <= *s && *s <= L'f') {
+                            value = *s - L'a' + 0xA;
+                        } else { // Oops, missing digit
+                            wb_wccat(&buf, L'?');
+                            break;
+                        }
+                        s++;
+                        if (L'0' <= *s && *s <= L'9')
+                            value = (value << 4) | (*s - L'0');
+                        else if (L'A' <= *s && *s <= L'F')
+                            value = (value << 4) | (*s - L'A' + 0xA);
+                        else if (L'a' <= *s && *s <= L'f')
+                            value = (value << 4) | (*s - L'a' + 0xA);
+                        else // Okay, no second digit
+                            goto only_one_xdigit;
+                        s++;
+only_one_xdigit:
+                        wb_wccat(&buf, (wchar_t) value);
+                        break;
+                    default:;
+                        int count = 0;
+                        value = 0;
+                        while (count < 3 && L'0' <= *s && *s <= L'7')
+                            value = (value << 3) | (*s - L'0'), s++, count++;
+                        if (count > 0 && (value & ~0xFF) == 0)
+                            wb_wccat(&buf, (wchar_t) value);
+                        else
+                            wb_wccat(&buf, L'?');
+                        break;
+                }
+                break;
+            default:
+                wb_wccat(&buf, *s);
+                s++;
+                break;
+        }
+    }
+}
+
 /* Backslashes characters in `s' that are contained in `t'.
  * Returns a newly-malloced wide string.
  * `t' may be NULL, in which case all the characters are backslashed. */
