@@ -41,7 +41,7 @@ extern _Bool is_interactive, is_interactive_now;
 extern _Bool shopt_cmdline, shopt_stdin;
 extern _Bool do_job_control, shopt_notify, shopt_notifyle,
        shopt_curasync, shopt_curbg, shopt_curstop;
-extern _Bool shopt_allexport, shopt_hashondef;
+extern _Bool shopt_allexport, shopt_hashondef, shopt_forlocal;
 extern _Bool shopt_errexit, shopt_errreturn, shopt_pipefail, shopt_unset,
        shopt_exec, shopt_ignoreeof, shopt_verbose, shopt_xtrace;
 extern _Bool shopt_traceall;
@@ -57,7 +57,8 @@ extern _Bool shopt_clobber;
 extern enum shopt_lineedit_T shopt_lineedit;
 extern enum shopt_yesnoauto_T shopt_le_convmeta;
 extern _Bool shopt_le_visiblebell, shopt_le_promptsp, shopt_le_alwaysrp,
-       shopt_le_predict, shopt_le_compdebug;
+    shopt_le_predict, shopt_le_predictempty, shopt_le_compdebug,
+    shopt_le_trimright;
 #endif
 
 /* Whether or not this shell process is doing job control right now. */
@@ -71,7 +72,7 @@ struct shell_invocation_T {
 };
 
 extern int parse_shell_options(int argc, void *const *argv,
-	struct shell_invocation_T *shell_invocation)
+        struct shell_invocation_T *shell_invocation)
     __attribute__((nonnull(2),warn_unused_result));
 #if YASH_ENABLE_LINEEDIT
 extern void set_lineedit_option(enum shopt_lineedit_T v);
@@ -101,4 +102,4 @@ extern const char set_help[], set_syntax[];
 #endif /* YASH_OPTION_H */
 
 
-/* vim: set ts=8 sts=4 sw=4 noet tw=80: */
+/* vim: set ts=8 sts=4 sw=4 et tw=80: */
