@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* path.c: filename-related utilities */
-/* (C) 2007-2020 magicant */
+/* (C) 2007-2025 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -1157,29 +1157,36 @@ static inline mode_t copy_other_mask(mode_t mode)
 /* The "cd" built-in, which accepts the following options:
  *  -L: don't resolve symbolic links (default)
  *  -P: resolve symbolic links
+ *  -e: fail if new $PWD value cannot be determined
  *  --default-directory=<dir>: go to <dir> when the operand is missing
  * -L and -P are mutually exclusive: the one specified last is used. */
 int cd_builtin(int argc, void **argv)
 {
-    bool logical = true;
+    bool logical = true, ensure_pwd = false;
     const wchar_t *newpwd = NULL;
 
     const struct xgetopt_T *opt;
     xoptind = 0;
     while ((opt = xgetopt(argv, cd_options, XGETOPT_DIGIT)) != NULL) {
         switch (opt->shortopt) {
-            case L'L':  logical = true;    break;
-            case L'P':  logical = false;   break;
-            case L'd':  newpwd = xoptarg;  break;
+            case L'L':  logical = true;     break;
+            case L'P':  logical = false;    break;
+            case L'e':  ensure_pwd = true;  break;
+            case L'd':  newpwd = xoptarg;   break;
 #if YASH_ENABLE_HELP
             case L'-':
                 return print_builtin_help(ARGV(0));
 #endif
             default:
-                return Exit_ERROR;
+                return 5;
         }
     }
 
+    if (logical && ensure_pwd) {
+        xerror(0, Ngt("the -e option requires the -P option"));
+        return 5;
+    }
+
     bool printnewdir = false;
     switch (argc - xoptind) {
         case 0:
@@ -1187,7 +1194,7 @@ int cd_builtin(int argc, void **argv)
                 newpwd = getvar(L VAR_HOME);
                 if (newpwd == NULL || newpwd[0] == L'\0') {
                     xerror(0, Ngt("$HOME is not set"));
-                    return Exit_FAILURE;
+                    return 4;
                 }
             }
             break;
@@ -1196,7 +1203,7 @@ int cd_builtin(int argc, void **argv)
                 newpwd = getvar(L VAR_OLDPWD);
                 if (newpwd == NULL || newpwd[0] == L'\0') {
                     xerror(0, Ngt("$OLDPWD is not set"));
-                    return Exit_FAILURE;
+                    return 4;
                 }
                 printnewdir = true;
             } else {
@@ -1204,9 +1211,10 @@ int cd_builtin(int argc, void **argv)
             }
             break;
         default:
-            return too_many_operands_error(1);
+            too_many_operands_error(1);
+            return 5;
     }
-    return change_directory(newpwd, printnewdir, logical);
+    return change_directory(newpwd, printnewdir, logical, ensure_pwd);
 }
 
 /* Changes the working directory to `newpwd'.
@@ -1214,19 +1222,22 @@ int cd_builtin(int argc, void **argv)
  * $PWD and $OLDPWD are set in this function.
  * If `printnewdir' is true or the new directory is found from $CDPATH, the new
  * directory is printed to the standard output.
- * Returns Exit_SUCCESS, Exit_FAILURE or Exit_ERROR. */
-int change_directory(const wchar_t *newpwd, bool printnewdir, bool logical)
+ * Returns the exit status to be returned from the built-in. */
+int change_directory(
+        const wchar_t *newpwd, bool printnewdir, bool logical, bool ensure_pwd)
 {
     const wchar_t *origpwd;
     xwcsbuf_T curpath;
     size_t curpathoffset = 0;
 
+    assert(!logical || !ensure_pwd);
+
     /* get the current value of $PWD as `origpwd' */
     origpwd = getvar(L VAR_PWD);
     if (origpwd == NULL || origpwd[0] != L'/') {
         if (origpwd == newpwd) {
             xerror(0, Ngt("$PWD has an invalid value"));
-            return Exit_FAILURE;
+            return 4;
         }
         /* we have to assure `origpwd != newpwd' because we're going to
          * re-assign $PWD */
@@ -1235,7 +1246,7 @@ int change_directory(const wchar_t *newpwd, bool printnewdir, bool logical)
         if (pwd == NULL) {
             if (logical) {
                 xerror(errno, Ngt("cannot determine the current directory"));
-                return Exit_FAILURE;
+                return 4;
             }
         } else {
             wchar_t *wpwd = realloc_mbstowcs(pwd);
@@ -1246,7 +1257,7 @@ int change_directory(const wchar_t *newpwd, bool printnewdir, bool logical)
                     logical = false, origpwd = NULL;
             } else {
                 xerror(EILSEQ, Ngt("cannot determine the current directory"));
-                return Exit_ERROR;
+                return 4;
             }
         }
     }
@@ -1272,7 +1283,7 @@ int change_directory(const wchar_t *newpwd, bool printnewdir, bool logical)
         if (mbsnewpwd == NULL) {
             wb_destroy(&curpath);
             xerror(EILSEQ, Ngt("unexpected error"));
-            return Exit_ERROR;
+            return 4;
         }
         char *const *cdpath = get_path_array(PA_CDPATH);
         char *path = which(mbsnewpwd,
@@ -1313,7 +1324,7 @@ step7:  /* ensure the value of `curpath' is an absolute path */
         wb_destroy(&curpath);
         if (canon == NULL) {
             xerror(ENOTDIR, Ngt("`%ls'"), newpwd);
-            return Exit_FAILURE;
+            return 3;
         }
         wb_initwith(&curpath, canon);
     }
@@ -1340,13 +1351,13 @@ step10:  /* do chdir */
         if (mbscurpath == NULL) {
             xerror(EILSEQ, Ngt("unexpected error"));
             wb_destroy(&curpath);
-            return Exit_ERROR;
+            return 2;
         }
         if (chdir(mbscurpath) < 0) {
             xerror(errno, Ngt("`%s'"), mbscurpath);
             free(mbscurpath);
             wb_destroy(&curpath);
-            return Exit_FAILURE;
+            return 2;
         }
         free(mbscurpath);
     }
@@ -1358,6 +1369,7 @@ step10:  /* do chdir */
 #endif
 
     /* set $OLDPWD and $PWD */
+    int result = 0;
     if (origpwd != NULL)
         set_variable(L VAR_OLDPWD, xwcsdup(origpwd), SCOPE_GLOBAL, false);
     if (logical) {
@@ -1377,12 +1389,16 @@ step10:  /* do chdir */
             wchar_t *wnewpwd = realloc_mbstowcs(mbsnewpwd);
             if (wnewpwd != NULL)
                 set_variable(L VAR_PWD, wnewpwd, SCOPE_GLOBAL, false);
+            else
+                result = ensure_pwd ? 1 : 0;
+        } else {
+            result = ensure_pwd ? 1 : 0;
         }
     }
     if (!posixly_correct)
         exec_variable_as_auxiliary_(VAR_YASH_AFTER_CD);
 
-    return Exit_SUCCESS;
+    return result;
 }
 
 /* Canonicalizes a pathname.
@@ -1493,7 +1509,7 @@ const char cd_help[] = Ngt(
 "change the working directory"
 );
 const char cd_syntax[] = Ngt(
-"\tcd [-L|-P] [directory]\n"
+"\tcd [-L|-P [-e]] [directory]\n"
 );
 #endif
 
